import logging

from .error_handler import ErrorHandler

logger = logging.getLogger('eipiphany.core.framework.base.DefaultErrorHandler')


class DefaultErrorHandler(ErrorHandler):

  def handle_exception(self, exchange):
    logger.error("Exception in route (" + exchange.get_header(ErrorHandler.EXCEPTION_CAUGHT) + ") " + exchange.get_header(ErrorHandler.EXCEPTION_CAUGHT_DETAIL))
    # print("Exception in route (" + exchange.get_header(ErrorHandler.EXCEPTION_CAUGHT) + ") " + exchange.get_header(ErrorHandler.EXCEPTION_CAUGHT_DETAIL))



import logging
from importlib import reload
from logging.handlers import QueueHandler


class SourceWrapper:

  def __init__(self, source, route, logging_queue, logging_level):
    self._source = source
    self.__route = route
    self.__logging_queue = logging_queue
    self.__logging_level = logging_level

  def wait_for_events(self):
    logging.shutdown()
    reload(logging)
    root = logging.getLogger()
    root.addHandler(QueueHandler(self.__logging_queue))
    root.setLevel(self.__logging_level)
    while True:
      exchange = self._source.wait_for_event()
      self.__route.run(exchange)

  def start(self):
    return self._source.start()


import abc

class ErrorHandler(metaclass=abc.ABCMeta):

  EXCEPTION_CAUGHT = 'ExceptionCaught'
  EXCEPTION_CAUGHT_DETAIL = 'ExceptionCaughtDetail'

  @abc.abstractmethod
  def handle_exception(self, exchange):
    pass




import copy
import logging
import traceback

from .exchange_handler import ExchangeHandler
from ..base.error_handler import ErrorHandler
from .source_wrapper import SourceWrapper

logger = logging.getLogger('eipiphany.core.framework.internal.Route')



class Route(object):

  def __init__(self, eip_context, endpoint_id, route_id):
    self.__source = eip_context.get_endpoint(endpoint_id).get_source()
    self._source_wrapper = SourceWrapper(self.__source, self, eip_context.logging_queue, eip_context.min_logging_level)
    self.__source.set_source_wrapper(self._source_wrapper)
    self._exchange_handlers = []
    self.__error_handler = None
    self.__route_id = route_id

  @property
  def route_id(self):
    return self.__route_id

  # todo move this to separate class
  def run(self, exchange):
    try:
      next_exchange = copy.deepcopy(exchange)
      for exchange_handler in self._exchange_handlers:
        if exchange_handler.processor:
          exchange_handler.processor.process(next_exchange)
        elif exchange_handler.endpoint:
          exchange_handler.endpoint.process(next_exchange, exchange_handler.endpoint_configuration)
        elif exchange_handler.filter:
          keep_going = exchange_handler.filter.filter(next_exchange)
          if not keep_going:
            break
        elif exchange_handler.aggregate:
          next_exchange = exchange_handler.aggregate.aggregate(next_exchange)
          if not next_exchange:
            break
        else:
          raise Exception("Internal error: invalid exchange handler")
      self.__source.event_success(exchange)
    except Exception as err:
      try:
        exchange.set_header(ErrorHandler.EXCEPTION_CAUGHT, str(err))
        exchange.set_header(ErrorHandler.EXCEPTION_CAUGHT_DETAIL, traceback.format_exc())
        self.__source.event_failure(err, exchange)
        self.__error_handler.handle_exception(exchange)
      except Exception as err2:
        # print("Exception in error handler: " + traceback.format_exc())
        logger.error("Exception in error handler", exc_info=err2)

  def process(self, processor):
    self._exchange_handlers.append(ExchangeHandler().set_processor(processor))
    return self

  def to(self, eip_context, endpoint_id, endpoint_configuration = None):
    self._exchange_handlers.append(
      ExchangeHandler()
      .set_endpoint(eip_context.get_endpoint(endpoint_id))
      .set_endpoint_configuration(endpoint_configuration))
    return self

  def filter(self, filter):
    self._exchange_handlers.append(ExchangeHandler().set_filter(filter))
    return self

  def error_handler(self, error_handler):
    self.__error_handler = error_handler
    return self

  def aggregate(self, aggregate):
    self._exchange_handlers.append(ExchangeHandler().set_aggregate(aggregate))
    return self

  def _set_default_error_handler(self, error_handler):
    if not self.__error_handler:
      self.__error_handler = error_handler

  def start(self):
    return self._source_wrapper.start()




import logging
import logging.config
import time
from importlib import reload
from logging.handlers import QueueHandler
from multiprocessing import Manager, Queue, Process

from .default_eip_context_termination import DefaultEipContextTermination
from .exchange_producer import ExchangeProducer
from ..internal.process_wrapper import ProcessWrapper

def _logging_process(queue, logging_config):
  logging.shutdown()
  reload(logging)
  logging.config.dictConfig(logging_config)
  while True:
    record = queue.get()
    if record is None:
      break
    logger = logging.getLogger(record.name)
    logger.handle(record)


_DEFAULT_LOGGING_CONFIG = {
  'version': 1,
  'disable_existing_loggers': False,
  'formatters': {
    'simple': {
      'format': '%(asctime)s - %(levelname)s - %(name)s - %(message)s',
      'datefmt': '%Y-%m-%d %H:%M:%S'
    }
  },
  'handlers': {
    'console': {
      'class': 'logging.StreamHandler',
      'formatter': 'simple',
      'stream': 'ext://sys.stdout'
    }
  },
  'loggers': {
    'eipiphany': {
      'level': 'INFO',
      'handlers': ['console'],
      'propagate': False
    },
  },
  'root': {
    'level': 'INFO',
    'handlers': ['console']
  }
}

_NAME_TO_LEVEL = {
  'CRITICAL': logging.CRITICAL,
  'FATAL': logging.FATAL,
  'ERROR': logging.ERROR,
  'WARN': logging.WARNING,
  'WARNING': logging.WARNING,
  'INFO': logging.INFO,
  'DEBUG': logging.DEBUG,
  'NOTSET': logging.NOTSET,
}

def _get_min_logging_level(logging_config):
  min_level = None
  loggers = logging_config.get('loggers')
  root = logging_config.get('root')
  if root and root.get('level'):
    level = _NAME_TO_LEVEL[root.get('level')]
    if min_level is None or level < min_level:
      min_level = level
  if loggers:
    for logger_name in loggers:
      logger = loggers[logger_name]
      if logger.get('level'):
        level = _NAME_TO_LEVEL[logger.get('level')]
        if min_level is None or level < min_level:
          min_level = level
  if min_level is None:
    min_level = logging.DEBUG
  return min_level



class EipContext(object):
  def __init__(self, termination=DefaultEipContextTermination(), logging_config=_DEFAULT_LOGGING_CONFIG):
    self.__logging_queue = Queue()
    self.__min_logging_level = _get_min_logging_level(logging_config)
    logging.shutdown()
    reload(logging)
    root = logging.getLogger()
    root.addHandler(QueueHandler(self.__logging_queue))
    root.setLevel(self.__min_logging_level)
    self.__logging_listener = Process(target=_logging_process, args=(self.__logging_queue, logging_config))
    self.__logging_listener.daemon = True
    self.__logging_listener.start()
    self.__manager = Manager()
    self._routes = []
    self.__processes = []
    self.__start_time = None
    self._termination = termination
    self.__endpoint_registry = {}
    self.__route_builders = []
    self.__exchange_producer = ExchangeProducer(self)

  def get_exchange_producer(self):
    return self.__exchange_producer

  def get_endpoint(self, endpoint_id):
    return self.__endpoint_registry[endpoint_id]

  def register_endpoint(self, endpoint):
    self._register_endpoint_internal(endpoint, False)

  def _register_endpoint_internal(self, endpoint, allow_override):
    prefix = endpoint.get_prefix()
    epid = prefix + ":" + endpoint.primary_id
    if not allow_override and self.__endpoint_registry.get(epid):
      raise Exception(epid + " is already registered")
    self.__endpoint_registry[epid] = endpoint

  def add_route_builder(self, route_builder):
    self.__route_builders.append(route_builder)
    return self

  def __terminate(self):
    for process in self.__processes:
      process.process.terminate()
      process.process.join()
      process.process.close()

  def start(self):
    self._start_internal(None)

  def _start_internal(self, after_start):
    for route_builder in self.__route_builders:
      route_builder.build(self)
      for route in route_builder.get_routes():
        route._set_default_error_handler(route_builder.error_handler)
        self._routes.append(route)
    self.__start_time = round(time.time() * 1000)
    for route in self._routes:
      for process in route.start():
        self.__processes.append(ProcessWrapper(process))
    if after_start:
      after_start.daemon = True
      after_start.start()
      self.__processes.append(ProcessWrapper(after_start, can_terminate=True))
    terminate = False
    while not terminate:
      time.sleep(1)
      terminate = self._termination.is_terminate(self)
    self.__terminate()

  @property
  def manager(self):
    return self.__manager

  @property
  def logging_queue(self):
    return self.__logging_queue

  @property
  def min_logging_level(self):
    return self.__min_logging_level

  @property
  def processes(self):
    return self.__processes

  @property
  def start_time(self):
    return self.__start_time

  def __enter__(self):
    self.__manager.__enter__()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.__manager.__exit__(exc_type, exc_val, exc_tb)
